import logging
import uuid
from typing import Callable, Optional

import pandas as pd

from autorag.utils.util import save_parquet_safe

logger = logging.getLogger("AutoRAG")


def make_single_content_qa(corpus_df: pd.DataFrame,
                           content_size: int,
                           qa_creation_func: Callable,
                           output_filepath: Optional[str] = None,
                           upsert: bool = False,
                           random_state: int = 42,
                           **kwargs) -> pd.DataFrame:
    """
    Make single content (single-hop, single-document) QA dataset using given qa_creation_func.
    It generates a single content QA dataset, which means its retrieval ground truth will be only one.
    It is the most basic form of QA dataset.

    :param corpus_df: The corpus dataframe to make QA dataset from.
    :param content_size: This function will generate QA dataset for the given number of contents.
    :param qa_creation_func: The function to create QA pairs.
        You can use like `generate_qa_llama_index` or `generate_qa_llama_index_by_ratio`.
        The input func must have `contents` parameter for the list of content string.
    :param output_filepath: Optional filepath to save the parquet file.
        If None, the function will return the processed_data as pd.DataFrame, but do not save as parquet.
        File directory must exist. File extension must be .parquet
    :param upsert: If true, the function will overwrite the existing file if it exists.
        Default is False.
    :param random_state: The random state for sampling corpus from the given corpus_df.
    :param kwargs: The keyword arguments for qa_creation_func.
    :return: QA dataset dataframe.
        You can save this as parquet file to use at AutoRAG.
    """
    assert content_size > 0, "content_size must be greater than 0."
    if content_size > len(corpus_df):
        logger.warning(f"content_size {content_size} is larger than the corpus size {len(corpus_df)}. "
                       "Setting content_size to the corpus size.")
        content_size = len(corpus_df)
    sampled_corpus = corpus_df.sample(n=content_size, random_state=random_state)
    sampled_corpus = sampled_corpus.reset_index(drop=True)

    qa = qa_creation_func(contents=sampled_corpus['contents'].tolist(), **kwargs)
    qa_data = pd.DataFrame({
        'qid': [str(uuid.uuid4()) for _ in range(len(qa))],
        'qa': qa,
        'retrieval_gt': sampled_corpus['doc_id'].tolist(),
    })
    qa_data = qa_data.explode('qa', ignore_index=True)

    def make_query_generation_gt(row):
        return row['qa']['query'], row['qa']['generation_gt']

    qa_data[['query', 'generation_gt']] = qa_data.apply(make_query_generation_gt, axis=1, result_type='expand')
    qa_data = qa_data.drop(columns=['qa'])

    qa_data['retrieval_gt'] = qa_data['retrieval_gt'].apply(lambda x: [[x]])
    qa_data['generation_gt'] = qa_data['generation_gt'].apply(lambda x: [x])

    if output_filepath is not None:
        save_parquet_safe(qa_data, output_filepath, upsert=upsert)

    return qa_data



import numpy as np


def calculate_cosine_similarity(a, b):
    dot_product = np.dot(a, b)
    norm_a = np.linalg.norm(a)
    norm_b = np.linalg.norm(b)
    similarity = dot_product / (norm_a * norm_b)
    return similarity



import pickle
import tempfile
from datetime import datetime

import pandas as pd
import pytest

from autorag.nodes.retrieval import bm25
from autorag.nodes.retrieval.bm25 import bm25_ingest, tokenize_ko_kiwi, tokenize_porter_stemmer, tokenize_space
from tests.autorag.nodes.retrieval.test_retrieval_base import (queries, project_dir, corpus_df, previous_result,
                                                               base_retrieval_test, base_retrieval_node_test)


@pytest.fixture
def ingested_bm25_path():
    with tempfile.NamedTemporaryFile(suffix='.pkl', mode='w+b') as path:
        bm25_ingest(path.name, corpus_df)
        yield path.name


def test_bm25_retrieval(ingested_bm25_path):
    with open(ingested_bm25_path, 'rb') as r:
        bm25_corpus = pickle.load(r)
    top_k = 3
    original_bm25 = bm25.__wrapped__
    id_result, score_result = original_bm25(queries, top_k=top_k, bm25_corpus=bm25_corpus)
    base_retrieval_test(id_result, score_result, top_k)


def test_bm25_node():
    result_df = bm25(project_dir=project_dir, previous_result=previous_result, top_k=4,
                     bm25_tokenizer='gpt2')
    base_retrieval_node_test(result_df)


def test_bm25_ingest(ingested_bm25_path):
    with open(ingested_bm25_path, 'rb') as r:
        corpus = pickle.load(r)
    assert set(corpus.keys()) == {'tokens', 'passage_id', 'tokenizer_name'}
    assert isinstance(corpus['tokens'], list)
    assert isinstance(corpus['passage_id'], list)
    assert isinstance(corpus['tokenizer_name'], str)
    assert corpus['tokenizer_name'] == 'porter_stemmer'
    assert len(corpus['tokens']) == len(corpus['passage_id']) == 5
    assert set(corpus['passage_id']) == {'doc1', 'doc2', 'doc3', 'doc4', 'doc5'}

    bm25_origin = bm25.__wrapped__
    top_k = 2
    id_result, score_result = bm25_origin([['What is test document?'], ['What is test document number 2?']],
                                          top_k=top_k, bm25_corpus=corpus)
    assert len(id_result) == len(score_result) == 2
    for id_list, score_list in zip(id_result, score_result):
        assert isinstance(id_list, list)
        assert isinstance(score_list, list)
        for _id in id_list:
            assert isinstance(_id, str)
            assert _id in ['doc1', 'doc2', 'doc3', 'doc4', 'doc5']


def test_duplicate_id_bm25_ingest(ingested_bm25_path):
    new_doc_id = ["doc4", "doc5", "doc6", "doc7", "doc8"]
    new_contents = ["This is a test document 4.", "This is a test document 5.", "This is a test document 6.",
                    "This is a test document 7.", "This is a test document 8."]
    new_metadata = [{'datetime': datetime.now()} for _ in range(5)]
    new_corpus_df = pd.DataFrame({"doc_id": new_doc_id, "contents": new_contents, "metadata": new_metadata})
    bm25_ingest(ingested_bm25_path, new_corpus_df)
    with open(ingested_bm25_path, 'rb') as r:
        corpus = pickle.load(r)
    assert len(corpus['tokens']) == len(corpus['passage_id']) == 8


def test_other_method_bm25(ingested_bm25_path):
    with open(ingested_bm25_path, 'rb') as r:
        corpus = pickle.load(r)
    bm25_origin = bm25.__wrapped__
    with pytest.raises(AssertionError):
        _, _ = bm25_origin([['What is test document?'], ['What is test document number 2?']],
                           top_k=1, bm25_corpus=corpus, bm25_tokenizer='gpt2')


def test_tokenize_ko_wiki():
    texts = [
        "안녕? 나는 혜인이야. 내가 비눗방울 만드는 방법을 알려줄께.",
        "너 정말 잘한다. 넌 정말 짱이야. 우리 친구할래?",
        "내 생일 파티에 너만 못 온 그날, 혜진이가 엄청 혼났던 그날, 지원이가 여친이랑 헤어진 그날",
    ]
    tokenized_list = tokenize_ko_kiwi(texts)
    assert len(tokenized_list) == len(texts)
    assert isinstance(tokenized_list[0], list)
    assert all(isinstance(x, str) for x in tokenized_list[0])


def test_tokenize_porter_stemmer():
    texts = [
        "The best baseball team in the world is Kia Tigers.",
        "And for a fortnight there, we were forever. Run into you sometimes, ask about the weather",
        "I walked through the door with you. The air was cold.",
    ]
    tokenized_list = tokenize_porter_stemmer(texts)
    assert len(tokenized_list) == len(texts)
    assert isinstance(tokenized_list[0], list)
    assert all(isinstance(x, str) for x in tokenized_list[0])


def test_tokenize_space():
    texts = [
        "내 생일 파티에 너만 못 온 그날, 혜진이가 엄청 혼났던 그날, 지원이가 여친이랑 헤어진 그날,,,, What's your ETA?  ",
        "  The best baseball team in the world is Kia Tigers. 최강 기아 타이거즈 최형우! 최형우!!! 기아의 해결사~   ",
        " You're my chemical hype boy (ah-ah) \n 내 지난날들은 눈 뜨면 잊는 꿈\nHype boy 너만 원해\nHype boy 내가 전해;",
    ]
    tokenized_list = tokenize_space(texts)
    assert len(tokenized_list) == len(texts)
    assert isinstance(tokenized_list[0], list)
    assert all(isinstance(x, str) for x in tokenized_list[0])



import os
import pathlib

import pandas as pd
import pytest

from autorag import dashboard
from autorag.dashboard import get_metric_values, make_trial_summary_md

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent
sample_project_dir = os.path.join(root_dir, 'resources', 'result_project')
sample_trial_dir = os.path.join(sample_project_dir, '0')


@pytest.fixture
def retrieval_summary_df():
    return pd.read_csv(os.path.join(sample_trial_dir, 'retrieve_node_line', 'retrieval', 'summary.csv'))


def test_get_metric_values(retrieval_summary_df):
    result_dict = get_metric_values(retrieval_summary_df)
    assert len(result_dict.keys()) == 3
    assert set(list(result_dict.keys())) == {'retrieval_f1', 'retrieval_recall', 'retrieval_precision'}
    assert result_dict['retrieval_recall'] == 1.0
    assert result_dict['retrieval_precision'] == 0.1


def test_make_trial_summary_md():
    md_text = make_trial_summary_md(sample_trial_dir)
    assert bool(md_text)


@pytest.mark.skip(reason="Can't stop this test on the github action or pytest cli setup")
def test_dashboard_run():
    dashboard.run(sample_trial_dir)



import importlib.resources
import logging
import os
import pathlib
import subprocess
from pathlib import Path
from typing import Optional

import click

from autorag import dashboard
from autorag.deploy import Runner
from autorag.deploy import extract_best_config as original_extract_best_config
from autorag.evaluator import Evaluator

logger = logging.getLogger("AutoRAG")


@click.group()
def cli():
    pass


@click.command()
@click.option('--config', '-c', help='Path to config yaml file. Must be yaml or yml file.', type=str)
@click.option('--qa_data_path', help='Path to QA dataset. Must be parquet file.', type=str)
@click.option('--corpus_data_path', help='Path to corpus dataset. Must be parquet file.', type=str)
@click.option('--project_dir', help='Path to project directory.', type=str, default=None)
def evaluate(config, qa_data_path, corpus_data_path, project_dir):
    if not config.endswith('.yaml') and not config.endswith('.yml'):
        raise ValueError(f"Config file {config} is not a parquet file.")
    if not os.path.exists(config):
        raise ValueError(f"Config file {config} does not exist.")
    evaluator = Evaluator(qa_data_path, corpus_data_path, project_dir=project_dir)
    evaluator.start_trial(config)
    logger.info('Evaluation complete.')


@click.command()
@click.option('--config_path', type=str, help='Path to extracted config yaml file.')
@click.option('--host', type=str, default='0.0.0.0', help='Host address')
@click.option('--port', type=int, default=8000, help='Port number')
@click.option('--project_dir', help='Path to project directory.', type=str, default=None)
def run_api(config_path, host, port, project_dir):
    runner = Runner.from_yaml(config_path, project_dir=project_dir)
    logger.info(f"Running API server at {host}:{port}...")
    runner.run_api_server(host, port)


@click.command()
@click.option('--yaml_path', type=click.Path(path_type=Path), help='Path to the YAML file.')
@click.option('--project_dir', type=click.Path(path_type=Path), help='Path to the project directory.')
@click.option('--trial_path', type=click.Path(path_type=Path), help='Path to the trial directory.')
def run_web(yaml_path: Optional[str], project_dir: Optional[str], trial_path: Optional[str]):
    try:
        with importlib.resources.path('autorag', 'web.py') as web_path:
            web_py_path = str(web_path)
    except ImportError:
        raise ImportError("Could not locate the web.py file within the autorag package."
                          " Please ensure that autorag is correctly installed.")

    if not yaml_path and not trial_path:
        raise ValueError("yaml_path or trial_path must be given.")
    elif yaml_path and trial_path:
        raise ValueError("yaml_path and trial_path cannot be given at the same time.")
    elif yaml_path and not project_dir:
        subprocess.run(['streamlit', 'run', web_py_path, '--', '--yaml_path', yaml_path])
    elif yaml_path and project_dir:
        subprocess.run(['streamlit', 'run', web_py_path, '--', '--yaml_path', yaml_path, '--project_dir', project_dir])
    elif trial_path:
        subprocess.run(['streamlit', 'run', web_py_path, '--', '--trial_path', trial_path])


@click.command()
@click.option('--trial_dir', type=click.Path(dir_okay=True, file_okay=False, exists=True), required=True)
def run_dashboard(trial_dir: str):
    dashboard.run(trial_dir)


@click.command()
@click.option('--trial_path', type=click.Path(), help='Path to the trial directory.')
@click.option('--output_path', type=click.Path(), help='Path to the output directory.'
                                                       ' Must be .yaml or .yml file.')
def extract_best_config(trial_path: str, output_path: str):
    original_extract_best_config(trial_path, output_path)


@click.command()
@click.option('--trial_path', help='Path to trial directory.', type=str)
def restart_evaluate(trial_path):
    if not os.path.exists(trial_path):
        raise ValueError(f"trial_path {trial_path} does not exist.")
    project_dir = pathlib.PurePath(trial_path).parent
    qa_data_path = os.path.join(project_dir, 'data', 'qa.parquet')
    corpus_data_path = os.path.join(project_dir, 'data', 'corpus.parquet')
    evaluator = Evaluator(qa_data_path, corpus_data_path, project_dir)
    evaluator.restart_trial(trial_path)
    logger.info('Evaluation complete.')


cli.add_command(evaluate, 'evaluate')
cli.add_command(run_api, 'run_api')
cli.add_command(run_web, 'run_web')
cli.add_command(run_dashboard, 'dashboard')
cli.add_command(extract_best_config, 'extract_best_config')
cli.add_command(restart_evaluate, 'restart_evaluate')

if __name__ == '__main__':
    cli()



from typing import List

import pandas as pd

from autorag.nodes.passageaugmenter.base import passage_augmenter_node


@passage_augmenter_node
def prev_next_augmenter(ids_list: List[List[str]],
                        corpus_df: pd.DataFrame,
                        num_passages: int = 1,
                        mode: str = 'both'
                        ) -> List[List[str]]:
    """
    Add passages before and/or after the retrieved passage.
    For more information, visit https://docs.llamaindex.ai/en/stable/examples/node_postprocessor/PrevNextPostprocessorDemo/.

    :param ids_list: The list of lists of ids retrieved
    :param corpus_df: The corpus dataframe
    :param num_passages: The number of passages to add before and after the retrieved passage
        Default is 1.
    :param mode: The mode of augmentation
        'prev': add passages before the retrieved passage
        'next': add passages after the retrieved passage
        'both': add passages before and after the retrieved passage
        Default is 'next'.
    :return: The list of lists of augmented ids
    """
    if mode not in ['prev', 'next', 'both']:
        raise ValueError(f"mode must be 'prev', 'next', or 'both', but got {mode}")

    augmented_ids = [(lambda ids: prev_next_augmenter_pure(ids, corpus_df, mode, num_passages))(ids) for ids in
                     ids_list]

    return augmented_ids


def prev_next_augmenter_pure(ids: List[str], corpus_df: pd.DataFrame, mode: str, num_passages: int):
    def fetch_id_sequence(start_id, key):
        sequence = []
        current_id = start_id
        for _ in range(num_passages):
            current_id = corpus_df.loc[corpus_df['doc_id'] == current_id]['metadata'].values[0].get(key)
            if current_id is None:
                break
            sequence.append(current_id)
        return sequence

    augmented_group = []
    for id_ in ids:
        current_ids = [id_]
        if mode in ['prev', 'both']:
            current_ids = fetch_id_sequence(id_, 'prev_id')[::-1] + current_ids
        if mode in ['next', 'both']:
            current_ids += fetch_id_sequence(id_, 'next_id')
        augmented_group.extend(current_ids)
    return augmented_group



import os
import pathlib
import subprocess
import tempfile
from distutils.dir_util import copy_tree

from click.testing import CliRunner

from autorag.cli import cli

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent
resource_dir = os.path.join(root_dir, 'resources')


def test_evaluator_cli():
    os.environ['BM25'] = 'bm25'
    with tempfile.TemporaryDirectory() as project_dir:
        result = subprocess.run(['autorag', 'evaluate', '--config', os.path.join(resource_dir, 'simple.yaml'),
                                 '--qa_data_path', os.path.join(resource_dir, 'qa_data_sample.parquet'),
                                 '--corpus_data_path', os.path.join(resource_dir, 'corpus_data_sample.parquet'),
                                 '--project_dir', project_dir])
        assert result.returncode == 0
        # check if the files are created
        assert os.path.exists(os.path.join(project_dir, '0'))
        assert os.path.exists(os.path.join(project_dir, 'data'))
        assert os.path.exists(os.path.join(project_dir, 'resources'))
        assert os.path.exists(os.path.join(project_dir, 'trial.json'))
        assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line'))
        assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'retrieval'))
        assert os.path.exists(os.path.join(project_dir, '0', 'retrieve_node_line', 'retrieval', '0.parquet'))


def test_run_api():
    runner = CliRunner()
    result = runner.invoke(cli, ['run_api', '--config_path', 'test/path/test.yaml',
                                 '--host', '0.0.0.0', '--port', '8080'])
    assert result.exit_code == 1  # it will occur error because I run this test with a wrong yaml path.
    # But it means that the command is working well. If not, it will occur exit_code 2.


def test_extract_best_config_cli():
    with tempfile.TemporaryDirectory() as project_dir:
        trial_path = os.path.join(resource_dir, 'result_project', '0')
        output_path = os.path.join(project_dir, 'best.yaml')
        subprocess.run(['autorag', 'extract_best_config', '--trial_path', trial_path, '--output_path', output_path])
        assert os.path.exists(output_path)


def test_restart_evaluate():
    with tempfile.TemporaryDirectory() as project_dir:
        original_path = os.path.join(resource_dir, 'result_project')
        copy_tree(original_path, project_dir)
        trial_path = os.path.join(project_dir, '1')
        subprocess.run(['autorag', 'restart_evaluate', '--trial_path', trial_path])
        assert os.path.exists(os.path.join(trial_path, 'summary.csv'))


def test_restart_evaluate_leads_start_evaluate():
    with tempfile.TemporaryDirectory() as project_dir:
        original_path = os.path.join(resource_dir, 'result_project')
        copy_tree(original_path, project_dir)
        trial_path = os.path.join(project_dir, '3')
        subprocess.run(['autorag', 'restart_evaluate', '--trial_path', trial_path])
        restart_path = os.path.join(project_dir, '4')
        assert os.path.exists(os.path.join(restart_path, 'summary.csv'))



import functools
import warnings
from typing import List, Callable, Any, Tuple, Optional, Union, Dict

import pandas as pd

from autorag.evaluate.metric import (retrieval_recall, retrieval_precision, retrieval_f1, retrieval_ndcg, retrieval_mrr,
                                     retrieval_map)
from autorag.evaluate.util import cast_metrics

RETRIEVAL_METRIC_FUNC_DICT = {func.__name__: func for func in
                              [retrieval_recall, retrieval_precision, retrieval_f1, retrieval_ndcg, retrieval_mrr,
                               retrieval_map]}
RETRIEVAL_NO_GT_METRIC_FUNC_DICT = {func.__name__: func for func in []}


def evaluate_retrieval(retrieval_gt: List[List[List[str]]], metrics: Union[List[str], List[Dict]],
                       queries: Optional[List[str]] = None, generation_gt: Optional[List[List[str]]] = None):
    def decorator_evaluate_retrieval(
            func: Callable[[Any], Tuple[List[List[str]], List[List[str]], List[List[float]]]]):
        """
        Decorator for evaluating retrieval results.
        You can use this decorator to any method that returns (contents, scores, ids),
        which is the output of conventional retrieval modules.

        :param func: Must return (contents, scores, ids)
        :return: wrapper function that returns pd.DataFrame, which is the evaluation result.
        """

        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> pd.DataFrame:
            contents, pred_ids, scores = func(*args, **kwargs)

            metric_scores = {}
            metric_names, metric_params = cast_metrics(metrics)

            for metric_name, metric_param in zip(metric_names, metric_params):
                if metric_name in RETRIEVAL_METRIC_FUNC_DICT:
                    metric_func = RETRIEVAL_METRIC_FUNC_DICT[metric_name]
                    metric_scores[metric_name] = metric_func(retrieval_gt=retrieval_gt, pred_ids=pred_ids,
                                                             **metric_param)
                elif metric_name in RETRIEVAL_NO_GT_METRIC_FUNC_DICT:
                    metric_func = RETRIEVAL_NO_GT_METRIC_FUNC_DICT[metric_name]
                    if queries is None:
                        raise ValueError(f"To using {metric_name}, you have to provide queries.")
                    if generation_gt is None:
                        raise ValueError(f"To using {metric_name}, you have to provide generation ground truth.")
                    metric_scores[metric_name] = metric_func(queries=queries, retrieved_contents=contents,
                                                             generation_gt=generation_gt, **metric_param)
                else:
                    warnings.warn(
                        f"metric {metric_name} is not in supported metrics: {RETRIEVAL_METRIC_FUNC_DICT.keys()}"
                        f" and {RETRIEVAL_NO_GT_METRIC_FUNC_DICT.keys()}"
                        f"{metric_name} will be ignored.")

            metric_result_df = pd.DataFrame(metric_scores)
            execution_result_df = pd.DataFrame({
                'retrieved_contents': contents,
                'retrieved_ids': pred_ids,
                'retrieve_scores': scores,
            })
            result_df = pd.concat([execution_result_df, metric_result_df], axis=1)
            return result_df

        return wrapper

    return decorator_evaluate_retrieval



from unittest.mock import patch

import openai.resources.chat
import pandas as pd

from autorag.nodes.generator import openai_llm
from tests.autorag.nodes.generator.test_generator_base import prompts, check_generated_texts, check_generated_tokens, \
    check_generated_log_probs
from tests.mock import mock_openai_chat_create


@patch.object(openai.resources.chat.completions.AsyncCompletions, 'create', mock_openai_chat_create)
def test_openai_llm():
    openai_original = openai_llm.__wrapped__
    model = "gpt-3.5-turbo"
    answers, tokens, log_probs = openai_original(prompts, model, batch=1, temperature=0.5, logprobs=False, n=3)
    check_generated_texts(answers)
    check_generated_tokens(tokens)
    check_generated_log_probs(log_probs)


@patch.object(openai.resources.chat.completions.AsyncCompletions, 'create', mock_openai_chat_create)
def test_openai_llm_node():
    previous_result = pd.DataFrame(
        {
            'prompts': prompts,
            'qid': ['id-1', 'id-2', 'id-3']
        })
    result_df = openai_llm(project_dir='.', previous_result=previous_result, llm='gpt-4o')
    check_generated_texts(result_df['generated_texts'].tolist())
    check_generated_tokens(result_df['generated_tokens'].tolist())
    check_generated_log_probs(result_df['generated_log_probs'].tolist())


@patch.object(openai.resources.chat.completions.AsyncCompletions, 'create', mock_openai_chat_create)
def test_openai_llm_truncate():
    openai_original = openai_llm.__wrapped__
    prompt = [f'havertz on the block and I am {i}th player on the Arsenal.' for i in range(50_000)]
    prompt = ' '.join(prompt)
    answers, tokens, log_probs = openai_original([prompt] * 3)
    check_generated_texts(answers)
    check_generated_tokens(tokens)
    check_generated_log_probs(log_probs)



from autorag.nodes.queryexpansion import hyde
from autorag.support import get_support_modules
from tests.autorag.nodes.queryexpansion.test_query_expansion_base import project_dir, previous_result, \
    base_query_expansion_node_test, ingested_vectordb_node

sample_query = ["How many members are in Newjeans?", "What is visconde structure?"]


def test_hyde():
    generator_func = get_support_modules('llama_index_llm')
    generator_params = {'llm': 'mock'}
    original_hyde = hyde.__wrapped__
    result = original_hyde(sample_query, generator_func, generator_params, prompt="")
    assert len(result[0]) == 1
    assert len(result) == 2


def test_hyde_node(ingested_vectordb_node):
    generator_dict = {
        'generator_module_type': 'llama_index_llm',
        'llm': 'mock'
    }
    result_df = hyde(project_dir=project_dir, previous_result=previous_result,
                     **generator_dict)
    base_query_expansion_node_test(result_df)



from typing import List

from autorag.nodes.queryexpansion.base import query_expansion_node


@query_expansion_node
def pass_query_expansion(queries: List[str]):
    """
    Do not perform query expansion.
    Return with the same queries.
    The dimension will be 2-d list, and the column name will be 'queries'.
    """
    return list(map(lambda x: [x], queries))



import asyncio
import os
import pickle
import re
from typing import List, Dict, Tuple, Callable, Union, Iterable

import numpy as np
import pandas as pd
from kiwipiepy import Kiwi, Token
from llama_index.core.indices.keyword_table.utils import simple_extract_keywords
from nltk import PorterStemmer
from rank_bm25 import BM25Okapi
from transformers import AutoTokenizer, PreTrainedTokenizerBase

from autorag.nodes.retrieval.base import retrieval_node, evenly_distribute_passages
from autorag.utils import validate_corpus_dataset
from autorag.utils.util import normalize_string


def tokenize_ko_kiwi(texts: List[str]) -> List[List[str]]:
    texts = list(map(lambda x: x.strip().lower(), texts))
    kiwi = Kiwi()
    tokenized_list: Iterable[List[Token]] = kiwi.tokenize(texts)
    return [list(map(lambda x: x.form, token_list)) for token_list in tokenized_list]


def tokenize_porter_stemmer(texts: List[str]) -> List[List[str]]:
    def tokenize_remove_stopword(text: str, stemmer) -> List[str]:
        text = text.lower()
        words = list(simple_extract_keywords(text))
        return [stemmer.stem(word) for word in words]

    stemmer = PorterStemmer()
    tokenized_list: List[List[str]] = list(map(lambda x: tokenize_remove_stopword(x, stemmer), texts))
    return tokenized_list


def tokenize_space(texts: List[str]) -> List[List[str]]:
    def tokenize_space_text(text: str) -> List[str]:
        text = normalize_string(text)
        return re.split(r'\s+', text.strip())

    return list(map(tokenize_space_text, texts))


BM25_TOKENIZER = {
    'porter_stemmer': tokenize_porter_stemmer,
    'ko_kiwi': tokenize_ko_kiwi,
    'space': tokenize_space,
}


@retrieval_node
def bm25(queries: List[List[str]], top_k: int, bm25_corpus: Dict, bm25_tokenizer: str = 'porter_stemmer') -> \
        Tuple[List[List[str]], List[List[float]]]:
    """
    BM25 retrieval function.
    You have to load a pickle file that is already ingested.

    :param queries: 2-d list of query strings.
        Each element of the list is a query strings of each row.
    :param top_k: The number of passages to be retrieved.
    :param bm25_corpus: A dictionary containing the bm25 corpus, which is doc_id from corpus and tokenized corpus.
        Its data structure looks like this:

        .. Code:: python

            {
                "tokens": [], # 2d list of tokens
                "passage_id": [], # 2d list of passage_id.
            }

    :param bm25_tokenizer: The tokenizer name that uses to the BM25.
        It supports 'porter_stemmer', 'ko_kiwi', and huggingface `AutoTokenizer`.
        You can pass huggingface tokenizer name.
        Default is porter_stemmer.
    :return: The 2-d list contains a list of passage ids that retrieved from bm25 and 2-d list of its scores.
        It will be a length of queries. And each element has a length of top_k.
    """
    # check if bm25_corpus is valid
    assert ("tokens" and "passage_id" in list(bm25_corpus.keys())), \
        "bm25_corpus must contain tokens and passage_id. Please check you ingested bm25 corpus correctly."
    tokenizer = select_bm25_tokenizer(bm25_tokenizer)
    assert bm25_corpus['tokenizer_name'] == bm25_tokenizer, \
        (f"The bm25 corpus tokenizer is {bm25_corpus['tokenizer_name']}, but your input is {bm25_tokenizer}. "
         f"You need to ingest again. Delete bm25 pkl file and re-ingest it.")
    bm25_instance = BM25Okapi(bm25_corpus["tokens"])
    # run async bm25_pure function
    tasks = [bm25_pure(input_queries, top_k, tokenizer, bm25_instance, bm25_corpus) for input_queries in queries]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(asyncio.gather(*tasks))
    id_result = list(map(lambda x: x[0], results))
    score_result = list(map(lambda x: x[1], results))
    return id_result, score_result


async def bm25_pure(queries: List[str], top_k: int, tokenizer, bm25_api: BM25Okapi, bm25_corpus: Dict) -> \
        Tuple[List[str], List[float]]:
    """
    Async BM25 retrieval function.
    Its usage is for async retrieval of bm25 row by row.

    :param queries: A list of query strings.
    :param top_k: The number of passages to be retrieved.
    :param tokenizer: A tokenizer that will be used to tokenize queries.
    :param bm25_api: A bm25 api instance that will be used to retrieve passages.
    :param bm25_corpus: A dictionary containing the bm25 corpus, which is doc_id from corpus and tokenized corpus.
        Its data structure looks like this:

        .. Code:: python

            {
                "tokens": [], # 2d list of tokens
                "passage_id": [], # 2d list of passage_id. Type must be str.
            }
    :return: The tuple contains a list of passage ids that retrieved from bm25 and its scores.
    """
    # I don't make queries operation to async, because queries length might be small, so it will occur overhead.
    if isinstance(tokenizer, PreTrainedTokenizerBase):
        tokenized_queries = tokenizer(queries).input_ids
    else:
        tokenized_queries = tokenizer(queries)
    id_result = []
    score_result = []
    for query in tokenized_queries:
        scores = bm25_api.get_scores(query)
        sorted_scores = sorted(scores, reverse=True)
        top_n_index = np.argsort(scores)[::-1][:top_k]
        ids = [bm25_corpus['passage_id'][i] for i in top_n_index]
        id_result.append(ids)
        score_result.append(sorted_scores[:top_k])

    # make a total result to top_k
    id_result, score_result = evenly_distribute_passages(id_result, score_result, top_k)
    # sort id_result and score_result by score
    result = [(_id, score) for score, _id in
              sorted(zip(score_result, id_result), key=lambda pair: pair[0], reverse=True)]
    id_result, score_result = zip(*result)
    return list(id_result), list(score_result)


def bm25_ingest(corpus_path: str, corpus_data: pd.DataFrame, bm25_tokenizer: str = 'porter_stemmer'):
    if not corpus_path.endswith('.pkl'):
        raise ValueError(f"Corpus path {corpus_path} is not a pickle file.")
    validate_corpus_dataset(corpus_data)
    ids = corpus_data['doc_id'].tolist()

    # Initialize bm25_corpus
    bm25_corpus = pd.DataFrame()

    # Load the BM25 corpus if it exists and get the passage ids
    if os.path.exists(corpus_path) and os.path.getsize(corpus_path) > 0:
        with open(corpus_path, 'rb') as r:
            corpus = pickle.load(r)
            bm25_corpus = pd.DataFrame.from_dict(corpus)
        duplicated_passage_rows = bm25_corpus[bm25_corpus['passage_id'].isin(ids)]
        new_passage = corpus_data[~corpus_data['doc_id'].isin(duplicated_passage_rows['passage_id'])]
    else:
        new_passage = corpus_data

    if not new_passage.empty:
        tokenizer = select_bm25_tokenizer(bm25_tokenizer)
        if isinstance(tokenizer, PreTrainedTokenizerBase):
            tokenized_corpus = tokenizer(new_passage['contents'].tolist()).input_ids
        else:
            tokenized_corpus = tokenizer(new_passage['contents'].tolist())
        new_bm25_corpus = pd.DataFrame({
            'tokens': tokenized_corpus,
            'passage_id': new_passage['doc_id'].tolist(),
        })

        if not bm25_corpus.empty:
            bm25_corpus_updated = pd.concat([bm25_corpus, new_bm25_corpus], ignore_index=True)
            bm25_dict = bm25_corpus_updated.to_dict('list')
        else:
            bm25_dict = new_bm25_corpus.to_dict('list')

        # add tokenizer name to bm25_dict
        bm25_dict['tokenizer_name'] = bm25_tokenizer

        with open(corpus_path, 'wb') as w:
            pickle.dump(bm25_dict, w)


def select_bm25_tokenizer(bm25_tokenizer: str) -> Callable[[str], List[Union[int, str]]]:
    if bm25_tokenizer in list(BM25_TOKENIZER.keys()):
        return BM25_TOKENIZER[bm25_tokenizer]

    return AutoTokenizer.from_pretrained(bm25_tokenizer, use_fast=False)



import time
from typing import Optional, Callable, Sequence, Any

import tiktoken
from llama_index.core.base.llms.types import ChatMessage, LLMMetadata, CompletionResponseGen
from llama_index.core.callbacks import CallbackManager
from llama_index.core.llms import CompletionResponse, CustomLLM
from llama_index.core.llms.callbacks import llm_completion_callback
from llama_index.core.types import PydanticProgramMode
from openai.types.chat import ChatCompletion, ChatCompletionMessage, ChatCompletionTokenLogprob
from openai.types.chat.chat_completion import Choice, ChoiceLogprobs


class MockLLM(CustomLLM):
    max_tokens: Optional[int]

    def __init__(
            self,
            max_tokens: Optional[int] = None,
            callback_manager: Optional[CallbackManager] = None,
            system_prompt: Optional[str] = None,
            messages_to_prompt: Optional[Callable[[Sequence[ChatMessage]], str]] = None,
            completion_to_prompt: Optional[Callable[[str], str]] = None,
            pydantic_program_mode: PydanticProgramMode = PydanticProgramMode.DEFAULT,
            **kwargs: Any,
    ) -> None:
        super().__init__(
            max_tokens=max_tokens,
            callback_manager=callback_manager,
            system_prompt=system_prompt,
            messages_to_prompt=messages_to_prompt,
            completion_to_prompt=completion_to_prompt,
            pydantic_program_mode=pydantic_program_mode,
        )

    @classmethod
    def class_name(cls) -> str:
        return "MockLLM"

    @property
    def metadata(self) -> LLMMetadata:
        return LLMMetadata(num_output=self.max_tokens or -1)

    def _generate_text(self, length: int) -> str:
        return " ".join(["text" for _ in range(length)])

    @llm_completion_callback()
    def complete(
            self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponse:
        response_text = (
            self._generate_text(self.max_tokens) if self.max_tokens else prompt
        )

        return CompletionResponse(
            text=response_text,
        )

    @llm_completion_callback()
    def stream_complete(
            self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponseGen:
        def gen_prompt() -> CompletionResponseGen:
            for ch in prompt:
                yield CompletionResponse(
                    text=prompt,
                    delta=ch,
                )

        def gen_response(max_tokens: int) -> CompletionResponseGen:
            for i in range(max_tokens):
                response_text = self._generate_text(i)
                yield CompletionResponse(
                    text=response_text,
                    delta="text ",
                )

        return gen_response(self.max_tokens) if self.max_tokens else gen_prompt()


async def mock_openai_chat_create(self, messages, model, **kwargs):
    tokenizer = tiktoken.encoding_for_model(model)
    tokens = tokenizer.encode(str(messages[0]['content']))
    if len(tokens) > 16_385:
        raise ValueError("The maximum number of tokens is 16_385")

    return ChatCompletion(
        id="test_id",
        choices=[
            Choice(
                finish_reason="stop",
                index=0,
                logprobs=ChoiceLogprobs(
                    content=[
                        ChatCompletionTokenLogprob(
                            token='Why',
                            logprob=-0.445,
                            top_logprobs=[],
                        ),
                        ChatCompletionTokenLogprob(
                            token=' not',
                            logprob=-0.223,
                            top_logprobs=[],
                        ),
                        ChatCompletionTokenLogprob(
                            token="<|end|>",
                            logprob=-0.443,
                            top_logprobs=[],
                        )
                    ]
                ),
                message=ChatCompletionMessage(
                    content="Why not",
                    role='assistant',
                )
            )
        ],
        created=int(time.time()),
        model=model,
        object='chat.completion',
    )



prompts = [
    "Who is the strongest Avenger?",
    "Who is the best soccer player in the world?",
    "Who is the president of the United States?",
]


def check_generated_texts(generated_texts):
    assert len(generated_texts) == len(prompts)
    assert isinstance(generated_texts[0], str)
    assert all(bool(text) is True for text in generated_texts)


def check_generated_tokens(tokens):
    assert len(tokens) == len(prompts)
    assert isinstance(tokens[0], list)
    assert isinstance(tokens[0][0], int)


def check_generated_log_probs(log_probs):
    assert len(log_probs) == len(prompts)
    assert isinstance(log_probs[0], list)
    assert isinstance(log_probs[0][0], float)



from typing import List, Tuple

import numpy as np
import pandas as pd
import torch
from tqdm import tqdm
from transformers import AutoModel, AutoTokenizer

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.utils.util import flatten_apply, sort_by_scores, select_top_k


@passage_reranker_node
def colbert_reranker(queries: List[str], contents_list: List[List[str]],
                     scores_list: List[List[float]], ids_list: List[List[str]],
                     top_k: int, batch: int = 64,
                     model_name: str = "colbert-ir/colbertv2.0",
                     ) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents with Colbert rerank models.
    You can get more information about a Colbert model at https://huggingface.co/colbert-ir/colbertv2.0.
    It uses BERT-based model, so recommend using CUDA gpu for faster reranking.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param batch: The number of queries to be processed in a batch
        Default is 64.
    :param model_name: The model name for Colbert rerank.
        You can choose colbert model for reranking.
        Default is "colbert-ir/colbertv2.0".
    :return: Tuple of lists containing the reranked contents, ids, and scores
    """
    device = "cuda" if torch.cuda.is_available() else "cpu"
    model = AutoModel.from_pretrained(model_name).to(device)
    tokenizer = AutoTokenizer.from_pretrained(model_name)

    # get query and content embeddings
    query_embedding_list = get_colbert_embedding_batch(queries, model, tokenizer, batch)
    content_embedding_list = flatten_apply(get_colbert_embedding_batch, contents_list, model=model, tokenizer=tokenizer,
                                           batch_size=batch)

    del model
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    df = pd.DataFrame({
        'ids': ids_list,
        'query_embedding': query_embedding_list,
        'contents': contents_list,
        'content_embedding': content_embedding_list,
    })
    temp_df = df.explode('content_embedding')
    temp_df['score'] = temp_df.apply(lambda x: get_colbert_score(x['query_embedding'], x['content_embedding']), axis=1)
    df['scores'] = temp_df.groupby(level=0, sort=False)['score'].apply(list).tolist()
    df[['contents', 'ids', 'scores']] = df.apply(sort_by_scores, axis=1, result_type='expand')
    results = select_top_k(df, ['contents', 'ids', 'scores'], top_k)

    return results['contents'].tolist(), results['ids'].tolist(), results['scores'].tolist()


def get_colbert_embedding_batch(input_strings: List[str],
                                model, tokenizer, batch_size: int) -> List[np.array]:

    encoding = tokenizer(input_strings, return_tensors="pt", padding=True, truncation=True,
                         max_length=model.config.max_position_embeddings)

    input_batches = slice_tokenizer_result(encoding, batch_size)
    result_embedding = []
    with torch.no_grad():
        for encoding_batch in tqdm(input_batches):
            result_embedding.append(model(**encoding_batch).last_hidden_state)
    total_tensor = torch.cat(result_embedding, dim=0)  # shape [batch_size, token_length, embedding_dim]
    tensor_results = list(total_tensor.chunk(total_tensor.size()[0]))

    if torch.cuda.is_available():
        return list(map(lambda x: x.detach().cpu().numpy(), tensor_results))
    else:
        return list(map(lambda x: x.detach().numpy(), tensor_results))


def slice_tokenizer_result(tokenizer_output, batch_size):
    input_ids_batches = slice_tensor(tokenizer_output["input_ids"], batch_size)
    attention_mask_batches = slice_tensor(tokenizer_output["attention_mask"], batch_size)
    token_type_ids_batches = slice_tensor(tokenizer_output.get("token_type_ids", None), batch_size)
    return [{"input_ids": input_ids, "attention_mask": attention_mask, "token_type_ids": token_type_ids}
            for input_ids, attention_mask, token_type_ids in
            zip(input_ids_batches, attention_mask_batches, token_type_ids_batches)]


def slice_tensor(input_tensor, batch_size):
    # Calculate the number of full batches
    num_full_batches = input_tensor.size(0) // batch_size

    # Slice the tensor into batches
    tensor_list = [input_tensor[i * batch_size:(i + 1) * batch_size] for i in range(num_full_batches)]

    # Handle the last batch if it's smaller than batch_size
    remainder = input_tensor.size(0) % batch_size
    if remainder:
        tensor_list.append(input_tensor[-remainder:])

    device = "cuda" if torch.cuda.is_available() else "cpu"
    tensor_list = list(map(lambda x: x.to(device), tensor_list))

    return tensor_list


def get_colbert_score(query_embedding: np.array, content_embedding: np.array) -> float:
    if query_embedding.ndim == 3 and content_embedding.ndim == 3:
        query_embedding = query_embedding.reshape(-1, query_embedding.shape[-1])
        content_embedding = content_embedding.reshape(-1, content_embedding.shape[-1])

    sim_matrix = np.dot(query_embedding, content_embedding.T) / (
            np.linalg.norm(query_embedding, axis=1)[:, np.newaxis] * np.linalg.norm(content_embedding, axis=1)
    )
    max_sim_scores = np.max(sim_matrix, axis=1)
    return float(np.mean(max_sim_scores))



import logging
import os
import pathlib
from copy import deepcopy
from typing import List, Callable, Dict, Optional

import pandas as pd

from autorag.nodes.retrieval.run import evaluate_retrieval_node
from autorag.strategy import measure_speed, filter_by_threshold, select_best
from autorag.support import get_support_modules
from autorag.utils.util import make_combinations, explode

logger = logging.getLogger("AutoRAG")


def run_query_expansion_node(modules: List[Callable],
                             module_params: List[Dict],
                             previous_result: pd.DataFrame,
                             node_line_dir: str,
                             strategies: Dict,
                             ) -> pd.DataFrame:
    """
    Run evaluation and select the best module among query expansion node results.
    Initially, retrieval is run using expanded_queries, the result of the query_expansion module.
    The retrieval module is run as a combination of the retrieval_modules in strategies.
    If there are multiple retrieval_modules, run them all and choose the best result.
    If there are no retrieval_modules, run them with the default of bm25.
    In this way, the best result is selected for each module, and then the best result is selected.

    :param modules: Query expansion modules to run.
    :param module_params: Query expansion module parameters.
    :param previous_result: Previous result dataframe.
        In this case, it would be qa data.
    :param node_line_dir: This node line's directory.
    :param strategies: Strategies for query expansion node.
    :return: The best result dataframe.
    """
    if not os.path.exists(node_line_dir):
        os.makedirs(node_line_dir)
    node_dir = os.path.join(node_line_dir, "query_expansion")
    if not os.path.exists(node_dir):
        os.makedirs(node_dir)
    project_dir = pathlib.PurePath(node_line_dir).parent.parent

    # run query expansion
    results, execution_times = zip(*map(lambda task: measure_speed(
        task[0], project_dir=project_dir, previous_result=previous_result, **task[1]), zip(modules, module_params)))
    average_times = list(map(lambda x: x / len(results[0]), execution_times))

    # save results to folder
    pseudo_module_params = deepcopy(module_params)
    for i, module_param in enumerate(pseudo_module_params):
        if 'prompt' in module_params:
            module_param['prompt'] = str(i)
    filepaths = list(map(lambda x: os.path.join(node_dir, f'{x}.parquet'), range(len(modules))))
    list(map(lambda x: x[0].to_parquet(x[1], index=False), zip(results, filepaths)))  # execute save to parquet
    filenames = list(map(lambda x: os.path.basename(x), filepaths))

    # make summary file
    summary_df = pd.DataFrame({
        'filename': filenames,
        'module_name': list(map(lambda module: module.__name__, modules)),
        'module_params': module_params,
        'execution_time': average_times,
    })

    # Run evaluation when there are more than one module.
    if len(modules) > 1:
        # pop general keys from strategies (e.g. metrics, speed_threshold)
        general_key = ['metrics', 'speed_threshold']
        general_strategy = dict(filter(lambda x: x[0] in general_key, strategies.items()))
        extra_strategy = dict(filter(lambda x: x[0] not in general_key, strategies.items()))

        # first, filter by threshold if it is enabled.
        if general_strategy.get('speed_threshold') is not None:
            results, filenames = filter_by_threshold(results, average_times, general_strategy['speed_threshold'],
                                                     filenames)

        # check metrics in strategy
        if general_strategy.get('metrics') is None:
            raise ValueError("You must at least one metrics for query expansion evaluation.")

        if extra_strategy.get('top_k') is None:
            extra_strategy['top_k'] = 10  # default value

        # get retrieval modules from strategy
        retrieval_callables, retrieval_params = make_retrieval_callable_params(extra_strategy)

        # get retrieval_gt
        retrieval_gt = pd.read_parquet(os.path.join(project_dir, "data", "qa.parquet"))['retrieval_gt'].tolist()

        # run evaluation
        evaluation_results = list(map(lambda result: evaluate_one_query_expansion_node(
            retrieval_callables, retrieval_params, result['queries'].tolist(), retrieval_gt,
            general_strategy['metrics'], project_dir, previous_result, strategies.get('strategy', 'mean')), results))

        evaluation_df = pd.DataFrame({
            'filename': filenames,
            **{f'query_expansion_{metric_name}': list(map(lambda x: x[metric_name].mean(), evaluation_results))
               for metric_name in general_strategy['metrics']}
        })
        summary_df = pd.merge(on='filename', left=summary_df, right=evaluation_df, how='left')

        best_result, best_filename = select_best(evaluation_results, general_strategy['metrics'], filenames,
                                                 strategies.get('strategy', 'mean'))
        # change metric name columns to query_expansion_metric_name
        best_result = best_result.rename(columns={
            metric_name: f'query_expansion_{metric_name}' for metric_name in strategies['metrics']})
        best_result = best_result.drop(columns=['retrieved_contents', 'retrieved_ids', 'retrieve_scores'])
    else:
        best_result, best_filename = results[0], filenames[0]
        best_result = pd.concat([previous_result, best_result], axis=1)

    # add 'is_best' column at summary file
    summary_df['is_best'] = summary_df['filename'] == best_filename

    # save files
    summary_df.to_csv(os.path.join(node_dir, "summary.csv"), index=False)
    best_result.to_parquet(os.path.join(node_dir, f"best_{os.path.splitext(best_filename)[0]}.parquet"), index=False)

    return best_result


def evaluate_one_query_expansion_node(retrieval_funcs: List[Callable],
                                      retrieval_params: List[Dict],
                                      expanded_queries: List[List[str]],
                                      retrieval_gt: List[List[str]],
                                      metrics: List[str],
                                      project_dir,
                                      previous_result: pd.DataFrame,
                                      strategy_name: str) -> pd.DataFrame:
    previous_result['queries'] = expanded_queries
    retrieval_results = list(map(lambda x: x[0](project_dir=project_dir, previous_result=previous_result, **x[1]),
                                 zip(retrieval_funcs, retrieval_params)))
    evaluation_results = list(map(lambda x: evaluate_retrieval_node(x, retrieval_gt, metrics,
                                                                    previous_result['query'].tolist(),
                                                                    previous_result['generation_gt'].tolist()),
                                  retrieval_results))
    best_result, _ = select_best(evaluation_results, metrics, strategy_name=strategy_name)
    best_result = pd.concat([previous_result, best_result], axis=1)
    return best_result


def make_retrieval_callable_params(strategy_dict: Dict):
    """
        strategy_dict looks like this:

        .. Code:: json

            {
                "metrics": ["retrieval_f1", "retrieval_recall"],
                "top_k": 50,
                "retrieval_modules": [
                  {"module_type": "bm25"},
                  {"module_type": "vectordb", "embedding_model": ["openai", "huggingface"]}
                ]
              }

    """
    node_dict = deepcopy(strategy_dict)
    retrieval_module_list: Optional[List[Dict]] = node_dict.pop('retrieval_modules', None)
    if retrieval_module_list is None:
        retrieval_module_list = [{
            'module_type': 'bm25',
        }]
    node_params = node_dict
    modules = list(map(lambda module_dict: get_support_modules(module_dict.pop('module_type')),
                       retrieval_module_list))
    param_combinations = list(map(lambda module_dict: make_combinations({**module_dict, **node_params}),
                                  retrieval_module_list))
    return explode(modules, param_combinations)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'AutoRAG'
copyright = '2024, Marker-Inc'
author = 'Marker-Inc'

with open('../../autorag/VERSION') as f:
    version = f.read().strip()

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "myst_parser",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_inline_tabs",
    "sphinxcontrib.googleanalytics",
    "sphinx_sitemap",
]
source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

templates_path = ['_templates']
exclude_patterns = []

googleanalytics_id = 'G-T8ZKQM6RT9'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'furo'
html_static_path = ['_static']
html_baseurl = 'https://marker-inc-korea.github.io/AutoRAG/'



import pandas as pd

from autorag.nodes.passagecompressor import pass_compressor

queries = [
    "What is the capital of France?",
    "What is the meaning of life?",
]
retrieved_contents = [
    ["Paris is the capital of France.", "France is a country in Europe.", "France is a member of the EU."],
    ["The meaning of life is 42.", "The meaning of life is to be happy.", "The meaning of life is to be kind."],
]


def test_pass_compressor():
    df = pd.DataFrame({
        'query': queries,
        'retrieved_contents': retrieved_contents,
        'retrieved_ids': [['id-1', 'id-2', 'id-3'], ['id-4', 'id-5', 'id-6']],
        'retrieve_scores': [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]],
    })
    result_df = pass_compressor("project_dir", df)
    assert result_df['retrieved_contents'].tolist() == retrieved_contents



import os

import pandas as pd


def validate_corpus(result_df: pd.DataFrame, length: int, parquet_filepath):
    assert isinstance(result_df, pd.DataFrame)
    assert len(result_df) == length
    assert 'doc_id' in result_df.columns
    assert 'contents' in result_df.columns
    assert 'metadata' in result_df.columns
    assert os.path.exists(parquet_filepath)

    assert ['test text'] * length == result_df['contents'].tolist()
    assert all(['last_modified_datetime' in metadata for metadata in result_df['metadata'].tolist()])
    assert all([isinstance(doc_id, str) for doc_id in result_df['doc_id'].tolist()])



import logging
import os
import pathlib
from typing import List, Callable, Dict, Tuple

import pandas as pd

from autorag.evaluate import evaluate_retrieval
from autorag.strategy import measure_speed, filter_by_threshold, select_best
from autorag.utils.util import load_summary_file

logger = logging.getLogger("AutoRAG")


def run_retrieval_node(modules: List[Callable],
                       module_params: List[Dict],
                       previous_result: pd.DataFrame,
                       node_line_dir: str,
                       strategies: Dict,
                       ) -> pd.DataFrame:
    """
    Run evaluation and select the best module among retrieval node results.

    :param modules: Retrieval modules to run.
    :param module_params: Retrieval module parameters.
    :param previous_result: Previous result dataframe.
        Could be query expansion's best result or qa data.
    :param node_line_dir: This node line's directory.
    :param strategies: Strategies for retrieval node.
    :return: The best result dataframe.
        It contains previous result columns and retrieval node's result columns.
    """
    if not os.path.exists(node_line_dir):
        os.makedirs(node_line_dir)
    project_dir = pathlib.PurePath(node_line_dir).parent.parent
    qa_df = pd.read_parquet(os.path.join(project_dir, "data", "qa.parquet"))
    retrieval_gt = qa_df['retrieval_gt'].tolist()
    retrieval_gt = [[[str(uuid) for uuid in sub_array] if sub_array.size > 0 else [] for sub_array in inner_array]
                    for inner_array in retrieval_gt]

    save_dir = os.path.join(node_line_dir, "retrieval")  # node name
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)

    def run_and_save(input_modules, input_module_params, filename_start: int):
        result, execution_times = zip(*map(lambda task: measure_speed(
            task[0], project_dir=project_dir, previous_result=previous_result, **task[1]),
                                           zip(input_modules, input_module_params)))
        average_times = list(map(lambda x: x / len(result[0]), execution_times))

        # run metrics before filtering
        if strategies.get('metrics') is None:
            raise ValueError("You must at least one metrics for retrieval evaluation.")
        result = list(map(lambda x: evaluate_retrieval_node(x, retrieval_gt, strategies.get('metrics'),
                                                            qa_df['query'].tolist(),
                                                            qa_df['generation_gt'].tolist()), result))

        # save results to folder
        filepaths = list(map(lambda x: os.path.join(save_dir, f'{x}.parquet'),
                             range(filename_start, filename_start + len(input_modules))))
        list(map(lambda x: x[0].to_parquet(x[1], index=False), zip(result, filepaths)))  # execute save to parquet
        filename_list = list(map(lambda x: os.path.basename(x), filepaths))

        summary_df = pd.DataFrame({
            'filename': filename_list,
            'module_name': list(map(lambda module: module.__name__, input_modules)),
            'module_params': input_module_params,
            'execution_time': average_times,
            **{metric: list(map(lambda result: result[metric].mean(), result)) for metric in
               strategies.get('metrics')},
        })
        summary_df.to_csv(os.path.join(save_dir, 'summary.csv'), index=False)
        return result, average_times, summary_df

    # run retrieval modules except hybrid
    hybrid_module_names = ['hybrid_rrf', 'hybrid_cc', 'hybrid_rsf', 'hybrid_dbsf']
    filename_first = 0
    if any([module.__name__ not in hybrid_module_names for module in modules]):
        non_hybrid_modules, non_hybrid_module_params = zip(*filter(lambda x: x[0].__name__ not in hybrid_module_names,
                                                                   zip(modules, module_params)))
        non_hybrid_results, non_hybrid_times, non_hybrid_summary_df = run_and_save(non_hybrid_modules,
                                                                                   non_hybrid_module_params, filename_first)
        filename_first += len(non_hybrid_modules)
    else:
        non_hybrid_results, non_hybrid_times, non_hybrid_summary_df = [], [], pd.DataFrame()

    if any([module.__name__ in hybrid_module_names for module in modules]):
        hybrid_modules, hybrid_module_params = zip(*filter(lambda x: x[0].__name__ in hybrid_module_names,
                                                           zip(modules, module_params)))
        if all(['target_module_params' in x for x in hybrid_module_params]):
            # If target_module_params are already given, run hybrid retrieval directly
            hybrid_results, hybrid_times, hybrid_summary_df = run_and_save(hybrid_modules, hybrid_module_params,
                                                                           filename_first)
            filename_first += len(hybrid_modules)
        else:
            target_modules = list(map(lambda x: x.pop('target_modules'), hybrid_module_params))
            target_filenames = list(map(lambda x: select_result_for_hybrid(save_dir, x), target_modules))
            ids_scores = list(map(lambda x: get_ids_and_scores(save_dir, x), target_filenames))
            target_module_params = list(map(lambda x: get_module_params(save_dir, x), target_filenames))
            hybrid_module_params = list(map(lambda x: {**x[0], **x[1]}, zip(hybrid_module_params, ids_scores)))
            real_hybrid_times = list(map(lambda filename: get_hybrid_execution_times(save_dir, filename), target_filenames))
            hybrid_results, hybrid_times, hybrid_summary_df = run_and_save(hybrid_modules, hybrid_module_params,
                                                                           filename_first)
            filename_first += len(hybrid_modules)
            hybrid_times = real_hybrid_times.copy()
            hybrid_summary_df['execution_time'] = hybrid_times
            hybrid_summary_df = edit_summary_df_params(hybrid_summary_df, target_modules, target_module_params)
    else:
        hybrid_results, hybrid_times, hybrid_summary_df = [], [], pd.DataFrame()

    summary = pd.concat([non_hybrid_summary_df, hybrid_summary_df], ignore_index=True)
    results = non_hybrid_results + hybrid_results
    average_times = non_hybrid_times + hybrid_times
    filenames = summary['filename'].tolist()

    # filter by strategies
    if strategies.get('speed_threshold') is not None:
        results, filenames = filter_by_threshold(results, average_times, strategies['speed_threshold'], filenames)
    selected_result, selected_filename = select_best(results, strategies.get('metrics'), filenames,
                                                     strategies.get('strategy', 'mean'))
    best_result = pd.concat([previous_result, selected_result], axis=1)

    # add summary.csv 'is_best' column
    summary['is_best'] = summary['filename'] == selected_filename

    # save the result files
    best_result.to_parquet(os.path.join(save_dir, f'best_{os.path.splitext(selected_filename)[0]}.parquet'),
                           index=False)
    summary.to_csv(os.path.join(save_dir, 'summary.csv'), index=False)
    return best_result


def evaluate_retrieval_node(result_df: pd.DataFrame, retrieval_gt, metrics,
                            queries: List[str], generation_gt: List[List[str]]) -> pd.DataFrame:
    """
    Evaluate retrieval node from retrieval node result dataframe.

    :param result_df: The result dataframe from a retrieval node.
    :param retrieval_gt: Ground truth for retrieval from qa dataset.
    :param metrics: Metric list from input strategies.
    :param queries: Query list from input strategies.
    :param generation_gt: Ground truth for generation from qa dataset.
    :return: Return result_df with metrics columns.
        The columns will be 'retrieved_contents', 'retrieved_ids', 'retrieve_scores', and metric names.
    """

    @evaluate_retrieval(retrieval_gt=retrieval_gt, metrics=metrics, queries=queries, generation_gt=generation_gt)
    def evaluate_this_module(df: pd.DataFrame):
        return df['retrieved_contents'].tolist(), df['retrieved_ids'].tolist(), df['retrieve_scores'].tolist()

    return evaluate_this_module(result_df)


def select_result_for_hybrid(node_dir: str, target_modules: Tuple) -> List[str]:
    """
    Get ids and scores of target_module from summary.csv and each result parquet file.

    :param node_dir: The directory of the node.
    :param target_modules: The name of the target modules.
    :return: A list of filenames.
    """

    def select_best_among_module(df: pd.DataFrame, module_name: str):
        modules_summary = df.loc[lambda row: row['module_name'] == module_name]
        if len(modules_summary) == 1:
            return modules_summary.iloc[0, :]
        elif len(modules_summary) <= 0:
            raise ValueError(f"module_name {module_name} does not exist in summary.csv. "
                             f"You must run {module_name} before running hybrid retrieval.")
        metrics = modules_summary.drop(columns=['filename', 'module_name', 'module_params', 'execution_time'])
        metric_average = metrics.mean(axis=1)
        metric_average = metric_average.reset_index(drop=True)
        max_idx = metric_average.idxmax()
        best_module = modules_summary.iloc[max_idx, :]
        return best_module

    summary_df = load_summary_file(os.path.join(node_dir, "summary.csv"))
    best_results = list(map(lambda module_name: select_best_among_module(summary_df, module_name), target_modules))
    best_filenames = list(map(lambda df: df['filename'], best_results))
    return best_filenames


def get_module_params(node_dir: str, filenames: List[str]) -> Tuple[Dict]:
    summary_df = load_summary_file(os.path.join(node_dir, "summary.csv"))
    best_results = summary_df[summary_df['filename'].isin(filenames)]
    module_params = best_results['module_params'].tolist()
    return tuple(module_params)


def edit_summary_df_params(summary_df: pd.DataFrame, target_modules, target_module_params) -> pd.DataFrame:
    def delete_ids_scores(x):
        del x['ids']
        del x['scores']
        return x

    summary_df['module_params'] = summary_df['module_params'].apply(delete_ids_scores)
    summary_df['new_params'] = [{'target_modules': x, 'target_module_params': y} for x, y in zip(target_modules, target_module_params)]
    summary_df['module_params'] = summary_df.apply(lambda row: {**row['module_params'], **row['new_params']}, axis=1)
    summary_df = summary_df.drop(columns=['new_params'])
    return summary_df


def get_ids_and_scores(node_dir: str, filenames: List[str]) -> Dict:
    best_results_df = list(map(lambda filename: pd.read_parquet(os.path.join(node_dir, filename)), filenames))
    ids = tuple(map(lambda df: df['retrieved_ids'].apply(list).tolist(), best_results_df))
    scores = tuple(map(lambda df: df['retrieve_scores'].apply(list).tolist(), best_results_df))
    return {
        'ids': ids,
        'scores': scores,
    }


def get_hybrid_execution_times(node_dir: str, filenames: List[str]) -> float:
    summary_df = load_summary_file(os.path.join(node_dir, "summary.csv"))
    best_results = summary_df[summary_df['filename'].isin(filenames)]
    execution_times = best_results['execution_time'].sum()
    return execution_times



import functools
import itertools
import math
from typing import List


def retrieval_metric(func):
    @functools.wraps(func)
    def wrapper(retrieval_gt: List[List[List[str]]], pred_ids: List[List[str]]) -> List[float]:
        results = []
        for gt, pred in zip(retrieval_gt, pred_ids):
            if gt == [[]] or any(bool(g_) is False for g in gt for g_ in g):
                results.append(None)
            else:
                results.append(func(gt, pred))
        return results

    return wrapper


@retrieval_metric
def retrieval_f1(gt: List[List[str]], pred: List[str]):
    """
    Compute f1 score for retrieval.

    :param gt: 2-d list of ground truth ids.
        It contains and/or connections between ids.
    :param pred: Prediction ids.
    :return: The f1 score.
    """
    recall_score = retrieval_recall.__wrapped__(gt, pred)
    precision_score = retrieval_precision.__wrapped__(gt, pred)
    if recall_score + precision_score == 0:
        return 0
    else:
        return 2 * (recall_score * precision_score) / (recall_score + precision_score)


@retrieval_metric
def retrieval_recall(gt: List[List[str]], pred: List[str]):
    gt_sets = [frozenset(g) for g in gt]
    pred_set = set(pred)
    hits = sum(any(pred_id in gt_set for pred_id in pred_set) for gt_set in gt_sets)
    recall = hits / len(gt) if len(gt) > 0 else 0.0
    return recall


@retrieval_metric
def retrieval_precision(gt: List[List[str]], pred: List[str]):
    gt_sets = [frozenset(g) for g in gt]
    pred_set = set(pred)
    hits = sum(any(pred_id in gt_set for gt_set in gt_sets) for pred_id in pred_set)
    precision = hits / len(pred) if len(pred) > 0 else 0.0
    return precision


@retrieval_metric
def retrieval_ndcg(gt: List[List[str]], pred: List[str]):
    gt_sets = [frozenset(g) for g in gt]
    pred_set = set(pred)
    relevance_scores = {pred_id: 1 if any(pred_id in gt_set for gt_set in gt_sets) else 0 for pred_id in pred_set}

    dcg = sum((2 ** relevance_scores[doc_id] - 1) / math.log2(i + 2) for i, doc_id in enumerate(pred))

    len_flatten_gt = len(list(itertools.chain.from_iterable(gt)))
    len_pred = len(pred)
    ideal_pred = [1] * min(len_flatten_gt, len_pred) + [0] * max(0, len_pred - len_flatten_gt)
    idcg = sum(relevance / math.log2(i + 2) for i, relevance in enumerate(ideal_pred))

    ndcg = dcg / idcg if idcg > 0 else 0
    return ndcg


@retrieval_metric
def retrieval_mrr(gt: List[List[str]], pred: List[str]) -> float:
    """
    Reciprocal Rank (RR) is the reciprocal of the rank of the first relevant item.
    Mean of RR in whole queries is MRR.
    """
    # Flatten the ground truth list of lists into a single set of relevant documents
    gt_sets = [frozenset(g) for g in gt]

    rr_list = []
    for gt_set in gt_sets:
        for i, pred_id in enumerate(pred):
            if pred_id in gt_set:
                rr_list.append(1.0 / (i + 1))
                break
    return sum(rr_list) / len(gt_sets) if rr_list else 0.0


@retrieval_metric
def retrieval_map(gt: List[List[str]], pred: List[str]) -> float:
    """
    Mean Average Precision (MAP) is the mean of Average Precision (AP) for all queries.
    """
    gt_sets = [frozenset(g) for g in gt]

    ap_list = []

    for gt_set in gt_sets:
        pred_hits = [1 if pred_id in gt_set else 0 for pred_id in pred]
        precision_list = [sum(pred_hits[:i + 1]) / (i + 1) for i, hit in enumerate(pred_hits) if hit == 1]
        ap_list.append(sum(precision_list) / len(precision_list) if precision_list else 0.0)

    return sum(ap_list) / len(gt_sets) if ap_list else 0.0



import time

import pandas as pd
import pytest

from autorag.strategy import (measure_speed, filter_by_threshold, select_best_average, select_best_rr,
                              select_normalize_mean)

sample_dfs = [
    pd.DataFrame(
        {'content': ['a', 'b', 'c'], 'retrieval_f1': [0.1, 0.2, 0.3], 'retrieval_recall': [0.1, 0.2, 0.3]}),
    pd.DataFrame(
        {'content': ['d', 'e', 'f'], 'retrieval_f1': [0.2, 0.3, 0.4], 'retrieval_recall': [0.2, 0.3, 0.4]}),
    pd.DataFrame(
        {'content': ['g', 'h', 'i'], 'retrieval_f1': [0.3, 0.4, 0.5], 'retrieval_recall': [0.3, 0.4, 0.5]}),
]
sample_metadatas = ['a', 'b', 'c']


def test_measure_speed():
    empty_result, five_seconds = measure_speed(time.sleep, 2)
    assert empty_result is None
    assert pytest.approx(2, 0.1) == five_seconds


def test_filter_by_threshold():
    results = [1, 2, 3, 4]
    values = [1, 2, 3, 4]
    threshold = 3
    filename = ['a', 'b', 'c', 'd']
    filtered_results, filtered_filenames = filter_by_threshold(results, values, threshold, filename)
    assert filtered_results == [1, 2, 3]
    assert filtered_filenames == ['a', 'b', 'c']

    filtered_results, _ = filter_by_threshold(results, values, threshold)
    assert filtered_results == [1, 2, 3]


def test_avoid_empty_result():
    results = [1, 2, 3, 4]
    values = [1, 2, 3, 4]
    threshold = 0.5
    filenames = ['a', 'b', 'c', 'd']
    filtered_results, filtered_filenames = filter_by_threshold(results, values, threshold, filenames)
    assert filtered_results == [1, 2, 3, 4]
    assert filtered_filenames == ['a', 'b', 'c', 'd']


def test_select_best_average():
    best_df, best_filename = select_best_average(sample_dfs, ['retrieval_f1', 'retrieval_recall'], sample_metadatas)
    assert best_df['content'].tolist() == ['g', 'h', 'i']
    assert best_df['retrieval_f1'].tolist() == [0.3, 0.4, 0.5]
    assert best_df['retrieval_recall'].tolist() == [0.3, 0.4, 0.5]
    assert best_filename == 'c'

    best_df, _ = select_best_average(sample_dfs, ['retrieval_f1', 'retrieval_recall'])
    assert best_df['content'].tolist() == ['g', 'h', 'i']


def test_select_best_rr():
    best_df, best_filename = select_best_rr(sample_dfs, ['retrieval_f1', 'retrieval_recall'], sample_metadatas)
    assert best_df['content'].tolist() == ['g', 'h', 'i']
    assert best_df['retrieval_f1'].tolist() == [0.3, 0.4, 0.5]
    assert best_df['retrieval_recall'].tolist() == [0.3, 0.4, 0.5]
    assert best_filename == 'c'

    best_df, _ = select_best_average(sample_dfs, ['retrieval_f1', 'retrieval_recall'])
    assert best_df['content'].tolist() == ['g', 'h', 'i']


def test_select_normalize_mean():
    best_df, best_filename = select_normalize_mean(sample_dfs, ['retrieval_f1', 'retrieval_recall'], sample_metadatas)
    assert best_df['content'].tolist() == ['g', 'h', 'i']
    assert best_df['retrieval_f1'].tolist() == [0.3, 0.4, 0.5]
    assert best_df['retrieval_recall'].tolist() == [0.3, 0.4, 0.5]
    assert best_filename == 'c'

    best_df, _ = select_best_average(sample_dfs, ['retrieval_f1', 'retrieval_recall'])
    assert best_df['content'].tolist() == ['g', 'h', 'i']



import pandas as pd
import pytest

from autorag.nodes.passagecompressor import longllmlingua
from tests.autorag.nodes.passagecompressor.test_base_passage_compressor import (queries, retrieved_contents,
                                                                                check_result, df)


@pytest.mark.skip(reason="This test needs CUDA enabled machine.")
def test_longllmlingua():
    result = longllmlingua.__wrapped__(queries, retrieved_contents, [], [])
    check_result(result)


@pytest.mark.skip(reason="This test needs CUDA enabled machine.")
def test_longllmlingua_node():
    result = longllmlingua(
        "project_dir",
        df,
        target_token=75,
    )
    assert isinstance(result, pd.DataFrame)
    contents = result['retrieved_contents'].tolist()
    assert isinstance(contents, list)
    assert len(contents) == len(queries)
    assert isinstance(contents[0], list)
    assert len(contents[0]) == 1
    assert isinstance(contents[0][0], str)
    assert bool(contents[0][0]) is True



from typing import List, Tuple

import numpy as np
import pandas as pd
import torch
from tqdm import tqdm
from transformers import AutoModelForSequenceClassification, AutoTokenizer

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.utils.util import make_batch, sort_by_scores, flatten_apply, select_top_k


@passage_reranker_node
def koreranker(queries: List[str], contents_list: List[List[str]],
               scores_list: List[List[float]], ids_list: List[List[str]],
               top_k: int, batch: int = 64) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents based on their relevance to a query using ko-reranker.
    ko-reranker is a reranker based on korean (https://huggingface.co/Dongjin-kr/ko-reranker).

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param batch: The number of queries to be processed in a batch
        Default is 64.
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    model_path = "Dongjin-kr/ko-reranker"
    tokenizer = AutoTokenizer.from_pretrained(model_path)
    model = AutoModelForSequenceClassification.from_pretrained(model_path)
    model.eval()

    # Determine the device to run the model on (GPU if available, otherwise CPU)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)

    nested_list = [list(map(lambda x: [query, x], content_list)) for query, content_list in zip(queries, contents_list)]
    scores_nps = flatten_apply(koreranker_run_model, nested_list, model=model, batch_size=batch,
                               tokenizer=tokenizer, device=device)

    rerank_scores = list(map(lambda scores: exp_normalize(np.array(scores)).astype(float), scores_nps))

    df = pd.DataFrame({
        'contents': contents_list,
        'ids': ids_list,
        'scores': rerank_scores,
    })
    df[['contents', 'ids', 'scores']] = df.apply(sort_by_scores, axis=1, result_type='expand')
    results = select_top_k(df, ['contents', 'ids', 'scores'], top_k)

    del model
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return results['contents'].tolist(), results['ids'].tolist(), results['scores'].tolist()


def koreranker_run_model(input_texts, model, tokenizer, device, batch_size: int):
    batch_input_texts = make_batch(input_texts, batch_size)
    results = []
    for batch_texts in tqdm(batch_input_texts):
        inputs = tokenizer(batch_texts, padding=True, truncation=True, return_tensors='pt', max_length=512)
        inputs = inputs.to(device)
        with torch.no_grad():
            scores = model(**inputs, return_dict=True).logits.view(-1, ).float()
            scores_np = scores.cpu().numpy()
            results.extend(scores_np)
    return results


def exp_normalize(x):
    b = x.max()
    y = np.exp(x - b)
    return y / y.sum()



import os
import pathlib

import pandas as pd
import pytest

from autorag.data.qacreation.ragas import generate_qa_ragas
from autorag.utils import validate_qa_dataset
from tests.delete_tests import is_github_action

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent
resource_dir = os.path.join(root_dir, "resources")
corpus_df = pd.read_parquet(os.path.join(resource_dir, 'corpus_data_sample.parquet'))


@pytest.mark.skipif(is_github_action(), reason="Skipping this test on GitHub Actions because it needs intensive "
                                               "OpenAI model requests.")
def test_generate_qa_ragas():
    qa_df = generate_qa_ragas(corpus_df, 3)
    assert len(qa_df) == 3
    validate_qa_dataset(qa_df)
    id_ = qa_df['retrieval_gt'].tolist()[0][0][0]
    assert isinstance(id_, str)
    assert id_ in corpus_df['doc_id'].tolist()



import math
from typing import Tuple, List

import pandas as pd

from autorag.evaluate import evaluate_retrieval

retrieval_gt = [[[f'test{i}-{j}'] for i in range(2)] for j in range(4)]
queries_example = ["Query 1", "Query 2", "Query 3", "Query 4"]
generation_gt_example = [['Jazz', 'Music'],
                         ['Havertz'],
                         ['William', 'Gamst'],
                         ['Kia', 'Tigers', 'V12']]
contents = [
    ['a', 'b', 'c', 'd'],
    ['better', 'bone', 'caleb', 'done'],
    ['ate', 'better', 'cortex', 'dad'],
    ['aim', 'bond', 'cane', 'dance'],
]
scores = [
    [0.1, 0.2, 0.3, 0.4],
    [0.4, 0.3, 0.2, 0.1],
    [0.5, 0.4, 0.3, 0.2],
    [0.6, 0.5, 0.4, 0.3],
]
ids = [
    [retrieval_gt[0][0][0], retrieval_gt[0][1][0], 'pred-0', 'pred-1'],
    ['pred-2', retrieval_gt[1][0][0], retrieval_gt[1][1][0], 'pred-3'],
    [f'pred-{i}' for i in range(4, 8)],
    [retrieval_gt[3][0][0], 'pred-8', 'pred-9', 'pred-10'],
]


@evaluate_retrieval(retrieval_gt=retrieval_gt,
                    metrics=['retrieval_recall', 'retrieval_precision', 'retrieval_f1'],
                    queries=queries_example, generation_gt=generation_gt_example)
def pseudo_retrieval() -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    return contents, ids, scores


@evaluate_retrieval(retrieval_gt=retrieval_gt,
                    metrics=[{'metric_name': 'retrieval_recall'}, {'metric_name': 'retrieval_f1'}],
                    queries=queries_example, generation_gt=generation_gt_example)
def pseudo_retrieval_dict_metric() -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    return contents, ids, scores


def test_evaluate_retrieval():
    result_df = pseudo_retrieval()
    assert isinstance(result_df, pd.DataFrame)
    assert len(result_df) == 4
    assert len(result_df.columns) == 6
    assert list(result_df.columns) == ['retrieved_contents', 'retrieved_ids', 'retrieve_scores', 'retrieval_recall',
                                       'retrieval_precision', 'retrieval_f1']
    recall = result_df['retrieval_recall'].tolist()
    recall_solution = [1, 1, 0, 0.5]
    for pred, sol in zip(recall, recall_solution):
        assert math.isclose(pred, sol, rel_tol=1e-4)

    precision = result_df['retrieval_precision'].tolist()
    precision_solution = [0.5, 0.5, 0, 0.25]
    for pred, sol in zip(precision, precision_solution):
        assert math.isclose(pred, sol, rel_tol=1e-4)

    f1 = result_df['retrieval_f1'].tolist()
    f1_solution = [1 / 1.5, 1 / 1.5, 0, 0.25 / 0.75]
    for pred, sol in zip(f1, f1_solution):
        assert math.isclose(pred, sol, rel_tol=1e-4)


def test_evaluate_retrieval_dict():
    result_df = pseudo_retrieval_dict_metric()
    assert isinstance(result_df, pd.DataFrame)
    assert len(result_df) == 4
    assert len(result_df.columns) == 5
    assert set(result_df.columns) == {'retrieved_contents', 'retrieved_ids', 'retrieve_scores', 'retrieval_recall',
                                      'retrieval_f1'}

    recall = result_df['retrieval_recall'].tolist()
    recall_solution = [1, 1, 0, 0.5]
    for pred, sol in zip(recall, recall_solution):
        assert math.isclose(pred, sol, rel_tol=1e-4)



import numpy as np

from autorag.nodes.passagefilter import threshold_cutoff
from tests.autorag.nodes.passagefilter.test_passage_filter_base import queries_example, contents_example, \
    scores_example, ids_example, base_passage_filter_test, project_dir, previous_result, base_passage_filter_node_test


def test_threshold_cutoff():
    original_cutoff = threshold_cutoff.__wrapped__
    contents, ids, scores = original_cutoff(
        queries_example, contents_example, scores_example, ids_example, threshold=0.6)
    base_passage_filter_test(contents, ids, scores)


def test_threshold_cutoff_reverse():
    original_cutoff = threshold_cutoff.__wrapped__
    contents, ids, scores = original_cutoff(
        queries_example, contents_example, scores_example, ids_example, threshold=0.4, reverse=True)
    base_passage_filter_test(contents, ids, scores)
    assert scores[0] == [0.1, 0.1]
    assert contents[0] == ["NomaDamas is Great Team", "havertz is suck at soccer"]


def test_threshold_cutoff_numpy():
    original_cutoff = threshold_cutoff.__wrapped__
    numpy_scores = np.array([[0.1, 0.8, 0.1, 0.5], [0.1, 0.2, 0.7, 0.3]])
    contents, ids, scores = original_cutoff(
        queries_example, contents_example, numpy_scores, ids_example, threshold=0.6)
    base_passage_filter_test(contents, ids, scores)
    assert scores == [[0.8], [0.7]]
    assert contents == [["Paris is the capital of France."], ["Newjeans has 5 members."]]


def test_threshold_cutoff_node():
    result_df = threshold_cutoff(
        project_dir=project_dir, previous_result=previous_result, threshold=0.9)
    base_passage_filter_node_test(result_df)



from itertools import chain
from typing import List, Tuple

import pandas as pd
import torch
from tqdm import tqdm
from transformers import T5Tokenizer, T5ForConditionalGeneration

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.utils.util import make_batch, sort_by_scores, flatten_apply, select_top_k

prediction_tokens = {
    'castorini/monot5-base-msmarco': ['▁false', '▁true'],
    'castorini/monot5-base-msmarco-10k': ['▁false', '▁true'],
    'castorini/monot5-large-msmarco': ['▁false', '▁true'],
    'castorini/monot5-large-msmarco-10k': ['▁false', '▁true'],
    'castorini/monot5-base-med-msmarco': ['▁false', '▁true'],
    'castorini/monot5-3b-med-msmarco': ['▁false', '▁true'],
    'castorini/monot5-3b-msmarco-10k': ['▁false', '▁true'],
    'unicamp-dl/mt5-base-en-msmarco': ['▁no', '▁yes'],
    'unicamp-dl/ptt5-base-pt-msmarco-10k-v2': ['▁não', '▁sim'],
    'unicamp-dl/ptt5-base-pt-msmarco-100k-v2': ['▁não', '▁sim'],
    'unicamp-dl/ptt5-base-en-pt-msmarco-100k-v2': ['▁não', '▁sim'],
    'unicamp-dl/mt5-base-en-pt-msmarco-v2': ['▁no', '▁yes'],
    'unicamp-dl/mt5-base-mmarco-v2': ['▁no', '▁yes'],
    'unicamp-dl/mt5-base-en-pt-msmarco-v1': ['▁no', '▁yes'],
    'unicamp-dl/mt5-base-mmarco-v1': ['▁no', '▁yes'],
    'unicamp-dl/ptt5-base-pt-msmarco-10k-v1': ['▁não', '▁sim'],
    'unicamp-dl/ptt5-base-pt-msmarco-100k-v1': ['▁não', '▁sim'],
    'unicamp-dl/ptt5-base-en-pt-msmarco-10k-v1': ['▁não', '▁sim'],
    'unicamp-dl/mt5-3B-mmarco-en-pt': ['▁', '▁true'],
    'unicamp-dl/mt5-13b-mmarco-100k': ['▁', '▁true'],
}


@passage_reranker_node
def monot5(queries: List[str], contents_list: List[List[str]],
           scores_list: List[List[float]], ids_list: List[List[str]],
           top_k: int, model_name: str = 'castorini/monot5-3b-msmarco-10k',
           batch: int = 64, ) \
        -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents based on their relevance to a query using MonoT5.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param model_name: The name of the MonoT5 model to use for reranking
        Note: default model name is 'castorini/monot5-3b-msmarco-10k'
            If there is a '/' in the model name parameter,
            when we create the file to store the results, the path will be twisted because of the '/'.
            Therefore, it will be received as '_' instead of '/'.
    :param batch: The number of queries to be processed in a batch
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    # replace '_' to '/'
    if '_' in model_name:
        model_name = model_name.replace('_', '/')
    # Load the tokenizer and model from the pre-trained MonoT5 model
    tokenizer = T5Tokenizer.from_pretrained(model_name)
    model = T5ForConditionalGeneration.from_pretrained(model_name).eval()
    # Retrieve the tokens used by the model to represent false and true predictions
    token_false, token_true = prediction_tokens[model_name]
    token_false_id = tokenizer.convert_tokens_to_ids(token_false)
    token_true_id = tokenizer.convert_tokens_to_ids(token_true)
    # Determine the device to run the model on (GPU if available, otherwise CPU)
    device = "cuda" if torch.cuda.is_available() else "cpu"
    model.to(device)
    nested_list = [list(map(lambda x: [f'Query: {query} Document: {x}'], content_list))
                   for query, content_list in zip(queries, contents_list)]

    rerank_scores = flatten_apply(monot5_run_model, nested_list, model=model, batch_size=batch, tokenizer=tokenizer,
                                  device=device, token_false_id=token_false_id, token_true_id=token_true_id)

    df = pd.DataFrame({
        'contents': contents_list,
        'ids': ids_list,
        'scores': rerank_scores,
    })
    df[['contents', 'ids', 'scores']] = df.apply(sort_by_scores, axis=1, result_type='expand')
    results = select_top_k(df, ['contents', 'ids', 'scores'], top_k)

    del model
    del tokenizer
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return results['contents'].tolist(), results['ids'].tolist(), results['scores'].tolist()


def monot5_run_model(input_texts, model, batch_size: int, tokenizer, device, token_false_id, token_true_id):
    batch_input_texts = make_batch(input_texts, batch_size)
    results = []
    for batch_texts in tqdm(batch_input_texts):
        flattened_batch_texts = list(chain.from_iterable(batch_texts))
        input_encodings = tokenizer(flattened_batch_texts, padding=True, truncation=True, max_length=512,
                                    return_tensors='pt').to(
            device)
        with torch.no_grad():
            outputs = model.generate(input_ids=input_encodings['input_ids'],
                                     attention_mask=input_encodings['attention_mask'],
                                     output_scores=True,
                                     return_dict_in_generate=True)

        # Extract logits for the 'false' and 'true' tokens from the model's output
        logits = outputs.scores[-1][:, [token_false_id, token_true_id]]
        # Calculate the softmax probability of the 'true' token
        probs = torch.nn.functional.softmax(logits, dim=-1)[:, 1]
        results.extend(probs.tolist())
    return results
